/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Copyright (c) 2001-2002  HASEGAWA Masafumi
 */

#include "ebi_book.h"
#include "ebi_error.h"
#include "ebi_textpage.h"
#include "ebi_imagepage.h"
#include "ebi_pagemaker.h"
#include "ebi_fontset.h"
#include "ebi_code.h"
#include "ebi_wave.h"

__EBI_BEGIN_NAMESPACE

Book::Books Book::s_books;

Book* Book::findBook (EB_Book * book)
{
  if (s_books.find (book) == s_books.end ())
    {
      return NULL;
    }
  return s_books[book];
}

Book::Book ()
{
  if (s_books.size () == 0)
    {
      eb_initialize_library ();
    }

  s_books[&m_book] = this;
  m_page = NULL;
  memset (&m_link, 0, sizeof (m_link));

  eb_initialize_book (&m_book);

  m_fontset = new Fontset (&m_book, EB_FONT_16);
}

Book::~Book ()
{
  if (m_fontset != NULL)
    {
      delete m_fontset;
    }

  close ();

  s_books.erase (&m_book);

  if (s_books.size () == 0)
    {
      eb_finalize_library ();
    }
}

bool Book::changeWork ()
{
  ERR err = chg_wrk (&m_link);
  if (err < 0)
    {
      Error::outputMessageBV ("chg_wrk", err);
      return false;
    }

  return true;
}

bool Book::open (const LINK & link)
{
  m_link = link;
  changeWork ();

  if (eb_bind (&m_book, "") == -1)
    {
      Error::outputMessageEB ("ҤǤޤ");
      return false;
    }

  eb_initialize_all_subbooks (&m_book);
  EB_Subbook_Code subbooks[EB_MAX_SUBBOOKS];
  int count;
  Error::s_error = eb_subbook_list (&m_book, subbooks, &count);
  if (Error::s_error != EB_SUCCESS)
    {
      Error::outputMessageEB ("ܤǤޤ");
      return false;
    }
  for (int i = 0; i < count; ++i)
    {
      m_subbooks.push_back (new Subbook (this, subbooks[i]));
    }

  m_appendix.open ();

  return true;
}

void Book::close ()
{
  m_appendix.close ();

  for (Subbooks::iterator si = m_subbooks.begin (); si != m_subbooks.end (); ++si)
    {
      if (*si != NULL)
	{
	  delete *si;
	}
    }
  m_subbooks.clear ();

  if (eb_is_bound (&m_book))
    {
      eb_finalize_book (&m_book);
    }
}

Hookset* Book::getHookset ()
{
  return &m_hookset;
}

Fontset* Book::getFontset ()
{
  return m_fontset;
}

bool Book::isCharCodeIso8859_1 ()
{
  EB_Character_Code charCode;
  eb_character_code (&m_book, &charCode);
  if (charCode == EB_CHARCODE_ISO8859_1)
    {
      return true;
    }
  return false;
}

Subbook* Book::getSubbook (EB_Subbook_Code subbook) const
{
  for (Subbooks::const_iterator si = m_subbooks.begin (); si != m_subbooks.end (); ++si)
    {
      if (subbook == (*si)->getSubbook ())
	{
	  return *si;
	}
    }
  return NULL;
}

Subbook* Book::getCurrSubbook ()
{
  EB_Subbook_Code subbook;
  eb_subbook (&m_book, &subbook);
  for (Subbooks::iterator si = m_subbooks.begin (); si != m_subbooks.end (); ++si)
    {
      if ((*si)->getSubbook () == subbook)
	{
	  return *si;
	}
    }
  
  return NULL;
}

bool Book::getSubbookTitle (EB_Subbook_Code subbook, String * str)
{
  char title[EB_MAX_TITLE_LENGTH + 1];
  Error::s_error = eb_subbook_title2 (&m_book, subbook, title);
  if (Error::s_error != EB_SUCCESS)
    {
      Error::outputMessageEB ("ȥǤޤ");
      return false;
    }

  if (isCharCodeIso8859_1 ())
    {
      Code::latin1ToTron (title, str);
    }
  else
    {
      Code::eucJpToTron (title, str);
    }

  return true;
}

bool Book::getSubbookTitlePage (TextPage* page)
{
  PageMaker pageMaker (page);
  m_hookset.setPageMaker (&pageMaker);

  bool bLatin1 = isCharCodeIso8859_1 ();
  for (Subbooks::iterator si = m_subbooks.begin (); si != m_subbooks.end (); ++si)
    {
      pageMaker.newLine ();

      EB_Position pos;
      bool result = (selectSubbook ((*si)->getSubbook ()) && getMenu (&pos));
      if (result)
	{
	  pageMaker.beginReference (*si);
	}

      String title = (*si)->getTitle ();
      if (bLatin1)
	{
	  pageMaker.putString (title, FONT_HALF);
	}
      else
	{
	  pageMaker.putString (title);
	}

      if (result)
	{
	  pageMaker.endReference (pos);
	}
    }

  m_hookset.setPageMaker (NULL);

  return true;
}

bool Book::selectSubbook (EB_Subbook_Code subbook)
{
  changeWork ();

  Error::s_error = eb_set_subbook (&m_book, subbook);
  if (Error::s_error != EB_SUCCESS)
    {
      Error::outputMessageEB ("ܤǤޤ");
      return false;
    }

  m_appendix.setSubbook (subbook);

  return true;
}

bool Book::readString (ReadFunction function, const EB_Position & pos, std::string* str)
{
  Error::s_error = eb_seek_text (&m_book, &pos);
  if (Error::s_error != EB_SUCCESS)
    {
      return false;
    }

  EB_Appendix *appendix = m_appendix.getAppendix ();
  EB_Hookset *hookset = m_hookset.getHookset ();
  char buffer[1024 + 1];
  int len;
  if (str != NULL)
    {
      str->erase ();
    }

  while ((Error::s_error = (*function) (&m_book, appendix, hookset, this, sizeof (buffer) - 1, buffer, &len)) == EB_SUCCESS && len > 0)
    {
      if (str != NULL)
	{
	  *str += buffer;
	}
    }
  if (Error::s_error != EB_SUCCESS)
    {
      return false;
    }

  return true;
}

bool Book::haveCopyright ()
{
  return eb_have_copyright (&m_book);
}

bool Book::getCopyright (EB_Position* pos)
{
  Error::s_error = eb_copyright (&m_book, pos);
  if (Error::s_error != EB_SUCCESS)
    {
      return false;
    }

  return true;
}

bool Book::haveMenu ()
{
  return eb_have_menu (&m_book);
}

bool Book::getMenu (EB_Position* pos)
{
  Error::s_error = eb_menu (&m_book, pos);
  if (Error::s_error != EB_SUCCESS)
    {
      return false;
    }

  return true;
}

bool Book::searchWord (const TC * word, SearchMethod search)
{
  int (*function) (EB_Book *, const char *);
  switch (search)
    {
    case SEARCH_WORD:
      function = eb_search_word;
      break;

    case SEARCH_ENDWORD:
      function = eb_search_endword;
      break;

    case SEARCH_EXACTWORD:
      function = eb_search_exactword;
      break;

    default:return false;
    }

  std::string str;
  if (isCharCodeIso8859_1 ())
    {
      Code::tronToLatin1 (word, &str);
    }
  else
    {
      Code::tronToEucJp (word, &str);
    }

  Error::s_error = (*function) (&m_book, str.c_str ());
  if (Error::s_error != EB_SUCCESS)
    {
      return false;
    }
  return true;
}

bool Book::getEntryPage (const TC * word, SearchMethod search, TextPage * page)
{
  if (!searchWord (word, search))
    {
      return false;
    }

  Subbook* subbook = getCurrSubbook ();
  String title = subbook->getTitle ();
  bool bLatin1 = isCharCodeIso8859_1 ();

  PageMaker pageMaker (page);
  m_hookset.setPageMaker (&pageMaker);
  const int c_size = 10;
  EB_Hit hit[c_size];
  std::string heading, prevHeading;
  EB_Position prevText = {0, 0};
  int count, entry = 0;
  while ((Error::s_error = eb_hit_list (&m_book, c_size, hit, &count)) == EB_SUCCESS && count > 0)
    {
      for (int i = 0; i < count; ++i)
	{
	  pageMaker.newLine ();
	  pageMaker.beginReference (subbook);
	  bool result = readString (eb_read_heading, hit[i].heading, &heading);
	  pageMaker.endReference (hit[i].text);

	  if (hit[i].text.page == prevText.page && hit[i].text.offset == prevText.offset && heading.compare (prevHeading) == 0)
	    {
	      pageMaker.undoLine ();
	      continue;
	    }

	  pageMaker.putChar (TK_KSP, FONT_FULL);
	  if (bLatin1)
	    {
	      pageMaker.putString (title, FONT_HALF);
	    }
	  else
	    {
	      pageMaker.putString (title);
	    }

	  if (!result)
	    {
	      count = -1;
	      break;
	    }

	  prevHeading = heading;
	  prevText = hit[i].text;
	  ++entry;

	  if (entry == MAX_ENTRY)
	    {
	      count = 0;
	      break;
	    }
	}
      if (count <= 0)
	{
	  break;
	}
    }
  m_hookset.setPageMaker (NULL);

  if (Error::s_error != EB_SUCCESS)
    {
      Error::outputMessageEB ("ҥåȤȥǤޤ");
      return false;
    }

  return true;
}

bool Book::getEntryPageAll (const TC * word, SearchMethod search, TextPage * page)
{
  for (Subbooks::iterator si = m_subbooks.begin (); si != m_subbooks.end (); ++si)
    {
      EB_Subbook_Code subbook = (*si)->getSubbook ();
      if (!selectSubbook (subbook))
	{
	  continue;
	}

      getEntryPage (word, search, page);
    }

  return true;
}

bool Book::getTextPage (const EB_Position & pos, TextPage * page)
{
  PageMaker pageMaker (page);
  m_hookset.setPageMaker (&pageMaker);

  pageMaker.newLine ();
  bool result = readString (eb_read_text, pos);

  m_hookset.setPageMaker (NULL);

  return result;
}

bool Book::getImagePage (const EB_Position & pos, ImageLink::ImageType type, int width, int height, ImagePage * page)
{
  switch (type)
    {
    case ImageLink::TYPE_MONO:
      Error::s_error = eb_set_binary_mono_graphic (&m_book, &pos, width, height);
      break;

    case ImageLink::TYPE_GRAY:
      Error::s_error = eb_set_binary_gray_graphic (&m_book, &pos, width, height);
      break;

    case ImageLink::TYPE_COLOR:
    case ImageLink::TYPE_JPEG:
      Error::s_error = eb_set_binary_color_graphic (&m_book, &pos);
      break;
    }
  if (Error::s_error != EB_SUCCESS)
    {
      Error::outputMessageEB ("ХʥǡǤޤ");
      return false;
    }

  char binary[EB_SIZE_PAGE];
  int length;
  while ((Error::s_error = eb_read_binary (&m_book, EB_SIZE_PAGE, binary, &length)) == EB_SUCCESS && length > 0)
    {
      if (!page->addBinary (binary, length))
	{
	  Error::outputMessage ("ХʥǡɲäǤޤ");
	  return false;
	}
    }
  if (Error::s_error != EB_SUCCESS)
    {
      Error::outputMessageEB ("ХʥǡǤޤ");
      return false;
    }
  if (!page->convertBinary (type))
    {
      Error::outputMessage ("ХʥǡѴǤޤ");
      return false;
    }

  return true;
}

bool Book::getSoundData (const EB_Position& pos, int size, Wave* wave)
{
  EB_Position end;
  end.page = pos.page + (size / EB_SIZE_PAGE);
  end.offset = pos.offset + (size % EB_SIZE_PAGE);
  if (EB_SIZE_PAGE <= end.offset)
    {
      end.offset -= EB_SIZE_PAGE;
      end.page++;
    }

  Error::s_error = eb_set_binary_wave (&m_book, &pos, &end);
  if (Error::s_error != EB_SUCCESS)
    {
      Error::outputMessageEB ("ХʥǡǤޤ");
      return false;
    }

  char binary[EB_SIZE_PAGE];
  int length;
  while ((Error::s_error = eb_read_binary (&m_book, EB_SIZE_PAGE, binary, &length)) == EB_SUCCESS && length > 0)
    {
      if (!wave->addBinary (binary, length))
	{
	  Error::outputMessage ("ХʥǡɲäǤޤ");
	  return false;
	}
    }
  if (Error::s_error != EB_SUCCESS)
    {
      Error::outputMessageEB ("ХʥǡǤޤ");
      return false;
    }

  return true;
}

#if 0
bool Book::getForwardPage (EB_Position* pos)
{
  Error::s_error = eb_forward_text (&m_book, m_appendix.getAppendix ());
  if (Error::s_error != EB_SUCCESS)
    {
      Error::outputMessageEB ("ʸƬФޤ");
      return false;
    }
  Error::s_error = eb_tell_text (&m_book, pos);
  if (Error::s_error != EB_SUCCESS)
    {
      Error::outputMessageEB ("֤߰ޤ");
      return false;
    }

  return true;
}

bool Book::getBackwardPage (EB_Position* pos)
{
  Error::s_error = eb_backward_text (&m_book, m_appendix.getAppendix ());
  if (Error::s_error != EB_SUCCESS)
    {
      Error::outputMessageEB ("ʸƬФޤ");
      return false;
    }
  // Ƭˤ뤳ȤϤʤȹͤơ2ƤӽФ
  Error::s_error = eb_backward_text (&m_book, m_appendix.getAppendix ());
  if (Error::s_error != EB_SUCCESS)
    {
      Error::outputMessageEB ("ʸƬФޤ");
    }
  Error::s_error = eb_tell_text (&m_book, pos);
  if (Error::s_error != EB_SUCCESS)
    {
      Error::outputMessageEB ("֤߰ޤ");
      return false;
    }

  return true;
}
#endif

__EBI_END_NAMESPACE
