/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Copyright (c) 2001  HASEGAWA Masafumi
 */

#include "ebi_wave.h"
#include "ebi_error.h"

__EBI_BEGIN_NAMESPACE

#pragma pack (1)
typedef struct
{
  UB chunkID[4]; 
  UW chunkSize;
}
ChunkHeader;

typedef struct
{
  ChunkHeader header; // "fmt "
  UH waveFormatType;
  UH channel;
  UW sampFreq;
  UW bytesPerSec;
  UH blockSizes;
  UH bitsPerSample;
}
FmtChunk;

typedef struct
{
  ChunkHeader header; // "data"
  char data[1];
}
DataChunk;

typedef struct
{
  ChunkHeader header; // "RIFF"
  UB formType[4]; // "WAVE"
  FmtChunk fmt;
  DataChunk data;
}
WaveFile;

Wave::Wave ()
{
  m_binary = NULL;
  m_length = 0;
}

Wave::~Wave ()
{
  clear ();
}

void Wave::clear ()
{
  if (m_binary != NULL)
    {
      free (m_binary);
      m_binary = NULL;
      m_length = 0;
    }
}

bool Wave::addBinary (const char* binary, UW length)
{
  if (m_binary == NULL)
    {
      m_binary = (char*) malloc (length);
      if (m_binary == NULL)
	{
	  return false;
	}
      memcpy (m_binary, binary, length);
      m_length = length;
    }
  else
    {
      char* temp = (char*) realloc (m_binary, m_length + length);
      if (temp == NULL)
	{
	  clear ();
	  return false;
	}
      m_binary = temp;
      memcpy (m_binary + m_length, binary, length);
      m_length += length;
    }
  
  return true;
}

bool Wave::isValid () const
{
  if (m_binary == NULL || m_length < sizeof (WaveFile))
    {
      return false;
    }

  const WaveFile* wave = (const WaveFile*) m_binary;
  if (memcmp(wave->header.chunkID, "RIFF", 4) != 0
      || memcmp(wave->formType, "WAVE", 4) != 0
      || memcmp(wave->fmt.header.chunkID, "fmt ", 4) != 0
      || memcmp(wave->data.header.chunkID, "data", 4) != 0)
    {
      return false;
    }
  if (wave->header.chunkSize + sizeof (ChunkHeader) > m_length)
    {
      return false;
    }

  return true;
}

bool Wave::getPcmMode (SDPcmMode* mode) const
{
  if (!isValid ())
    {
      return false;
    }
  const WaveFile* wave = (const WaveFile*) m_binary;

  memset (mode, 0, sizeof (SDPcmMode));

  switch (wave->fmt.waveFormatType)
    {
    case 0x01: // ˥PCM
      switch (wave->fmt.bitsPerSample)
	{
	case 8:
	  mode->datafmt = PCMFmt8;
	  break;
	  
	case 16:
	  mode->datafmt = PCMFmt16;
	  break;

	default:
	  return false;
	}
      break;

    case 0x07: // -Law
      switch (wave->fmt.bitsPerSample)
	{
	case 8:
	  mode->datafmt = PCMFmt8ulaw;
	  break;
	  
	default:
	  return false;
	}
      break;

    default:
      return false;
    }

  switch (wave->fmt.channel)
    {
    case 1: // Mono
      mode->stereo = False;
      break;

    case 2: // Streo
      mode->stereo = True;
      break;

    default:
      return false;
    }

  switch (wave->fmt.sampFreq)
    {
    case 8000: // 8.0KHz
      mode->sampling = PCMST8K;
      break;

    case 11025: // 11.025KHz
      mode->sampling = PCMST11K;
      break;

    case 16000: // 16.0KHz
      mode->sampling = PCMST16K;
      break;

    case 22050: // 22.05KHz
      mode->sampling = PCMST22K;
      break;

    case 44100: // 44.1KHz
      mode->sampling = PCMST44K;
      break;

    case 48000: // 48.0KHz
      mode->sampling = PCMST48K;
      break;

    default:
      return false;
    }
  
  return true;
}
  
const char* Wave::getData (int* length) const
{
  if (!isValid ())
    {
      return NULL;
    }
  const WaveFile* wave = (const WaveFile*) m_binary;

  if (length != NULL)
    {
      *length = wave->data.header.chunkSize;
    }
  return wave->data.data;
}

__EBI_END_NAMESPACE
